#!/usr/bin/perl -w

use Jabber::Connection;
use Jabber::NodeFactory;
use Jabber::NS qw(:all);
use BerkeleyDB;

use strict;
use open ':utf8';

use constant SERVER		=> 'jabber.linux.it';
use constant JID		=> 'presence';
use constant PASSWORD	=> 'xxxxxxxx';
use constant DATADIR	=> '/home/jabber/presence/db';

use constant VERBOSE	=> 2;

my @blacklist = qw(
);

my $DB;

db_init();

my %blacklisted = map { $_ => undef } @blacklist;
undef @blacklist;

my $c = new Jabber::Connection(
	server		=> SERVER,
	log			=> 0,
	debug		=> 0,
);

log3("Making connection to Jabber server");
$c->connect or die "oops: ".$c->lastError;
$SIG{HUP} = $SIG{TERM} = $SIG{INT} = \&stop;

$c->register_handler('presence' => \&presence_handler);

$c->register_handler('iq' => \&iq_version);
$c->register_handler('iq' => \&iq_notimpl);

$c->auth(JID, PASSWORD, 'webpresence');

my $iq = new Jabber::NodeFactory::Node('iq');
$iq->attr('type', IQ_GET);
$iq->insertTag('query', NS_ROSTER);
$c->send($iq);
$c->send('<presence/>');

log3("Entering main loop");
$c->start;
exit 1;

##############################################################################
sub presence_handler {
	my ($node) = @_;

	log3("[presence_handler]");
	my $from = $node->attr('from') || return '';
	my $to = $node->attr('to') || return '';
	my $type = $node->attr('type') || '';
	my $show = $node->getTag('show') || '';
	$show = $show->data if $show;
	log3("--> presence from $from to $to ($type;$show)");

	my $fid = lc $from;
	$fid =~ s#\/.*$##;		# remove any resource suffix from JID
	my ($tid, $tsrv, $tres) = $to =~ m#^(.+)@(.+)(?:/(.+))?$#;
	$tres ||= '';

	if (not $type or $type eq 'available') {
		if (exists $blacklisted{$fid}) {
			log2("$fid is blacklisted, ignored");
			return r_HANDLED;
		}
		log3("$fid logged in");
		db_put($fid, time . ' ONLINE' . ($show ? " $show" : ''));
		my $reply = $node->reply;			# send our presence
		$c->send($reply);
	} elsif ($type eq 'unavailable') {
		log3("$fid logged out ($tres) ($to)");
		db_put($fid, time . ' OFFLINE');
	} elsif ($type eq 'subscribed') {		# subscription confirmation
	} elsif ($type eq 'subscribe') {
		log2("$fid requests subscription");
		my $reply = $node->reply;
		$reply->attr(type => 'subscribed');	# accept
		$c->send($reply);
		$reply->attr(type => 'subscribe');	# request subscription to them
		$c->send($reply);
	} elsif ($type eq 'unsubscribed') {		# unsubscription confirmation
	} elsif ($type eq 'unsubscribe') {
		log2("$fid requests unsubscription");
		db_del($fid);
		my $reply = $node->reply;
		$reply->attr(type => 'unsubscribed');	# acknowledge
		$c->send($reply);
		$reply->attr(type => 'unsubscribe');	# request unsubscr. from them
		$c->send($reply);
	} elsif ($type eq 'error') {
		my $error = $node->getTag('error') || '';
		my $errordata = '';
		my $errorcode = '';
		if ($error) {
			$errordata = $error->data;
			$errorcode = $error->attr('code');
		}
		log2("$fid reported an error: $errordata ($errorcode)");
		# XXX remove the user from the database?
	} else {
		log1("$fid sent unknown presence $type to $tid");
		use Data::Dumper; print STDERR Data::Dumper->Dumpxs([$node], ['node']);
		return '';
	}

	return r_HANDLED;
}

##############################################################################
sub iq_version {
	my ($node) = @_;

	log3("[iq_version]");
	my $query = $node->getTag('', NS_VERSION);
	return if not $query and $node->attr('type', IQ_GET);

	my $reply = $node->reply;
	$reply->attr('code', IQ_RESULT);
	$query->insertTag('display')->data($^O);
	$query->insertTag('name')->data('Presence Agent');
	$query->insertTag('version')->data('0.2');
	$c->send($reply);

	return r_HANDLED;
}

sub iq_notimpl {
	my ($node) = @_;

	log1("[iq_notimpl]");
	my $from = $node->attr('from');
	my $to = $node->attr('to') || return '';
	my $type = $node->attr('type') || '';
	log3("--> IGNORED IQ from $from to $to ($type)");
	return r_HANDLED if $type eq 'get';
	use Data::Dumper; print STDERR Data::Dumper->Dumpxs([$node], ['node']);
	print STDERR "======================================================\n";
	return '';
	return r_HANDLED;

	my $reply = $node->reply;
	$reply->attr('type', IQ_ERROR);
	my $error = $reply->insertTag('error');
	$error->attr('code', '501');
	$error->data('Not Implemented');
	$c->send($reply);

	return r_HANDLED;
}

##############################################################################
# exit the program gracefully
sub stop {
	log2("Exiting because of signal");

	exit 0 if not $c;

	$c->process(0);
	$c->disconnect;
	exit 0;
}

sub Jabber::NodeFactory::Node::reply {
	my ($self, $node) = @_;

	my $reply = new Jabber::NodeFactory::Node($self->name);
	my $id = $self->attr('id');
	$reply->attr('id', $id) if defined $id;
	#$reply->attr('to' => $self->attr('from'));
	my $from = Jabber::NodeFactory::Node::_encode($self->attr('from'));
	$reply->attr('to' => $from);
	$reply->attr('from' => $self->attr('to'));
	return $reply;
}


sub log1 {
	return unless VERBOSE >= 1;
	print STDERR "WARN: $_[0]\n";
}

sub log2 {
	return unless VERBOSE >= 2;
	print STDERR "INFO: $_[0]\n";
}

sub log3 {
	return unless VERBOSE >= 3;
	print STDERR "DBUG: $_[0]\n";
}


sub db_init {
	my $env = BerkeleyDB::Env->new(
		-Flags	=> DB_CREATE|DB_INIT_MPOOL|DB_INIT_CDB,
		-Home	=> DATADIR,
		-ErrFile => DATADIR . '/err.log',
		-Verbose => 1|2|4|8,
	);
	die "Cannot create DB environment: $BerkeleyDB::Error" if not defined $env;

	$DB = BerkeleyDB::Hash->new(
		-Env => $env,
		-Flags => DB_CREATE,
		-Filename => 'presence.db',
	);
	die "Cannot open DB: $BerkeleyDB::Error" if not defined $DB;
}

sub db_put {
	my ($key, $value) = @_;

	my $r = $DB->db_put($key, $value);
	die "db_put($key, $value): $r" if $r;
}

sub db_del {
	my ($key) = @_;

	my $r = $DB->db_del($key);
	die "db_del($key): $r" if $r and $r !~ /^DB_NOTFOUND/;
}

sub db_get {
	my ($key) = @_;
	my $value;

	my $r = $DB->db_get($key, $value);
	die "db_get($key): $r" if $r and $r !~ /^DB_NOTFOUND/;
	return undef if $r;
	return $value;
}

